@@ -27,6 +27,15 @@ sub transform {
         $targetpunc = $1;
     }
 
+    if ($sourcepunc eq $targetpunc && $targetpunc eq '') {
+        # try to get rid of the trailing tags and whitespace to get the actual punctuation
+        my $filtered_target = $target;
+        $filtered_target =~ s/\s*<[^<>]+>\s*$//s;
+        if ($filtered_target ne $target) {
+            return $self->transform($s, $filtered_target, $lang);
+        }
+    }
+
     if ($sourcepunc ne $targetpunc) {
         # In Japanese, replace certain punctuation symbols
         if ($lang =~ m/^(ja|ja-.*)$/) {
