@@ -3,6 +3,8 @@ use parent Serge::Engine::Plugin::Base::Parser;
 
 use strict;
 
+our $VERSION = 2;
+
 sub name {
     return 'Generic JavaScript object parser plugin';
 }
@@ -33,16 +35,52 @@ sub parse_callback {
     push @hint, $key if $key ne '' && $key ne $string;
     push @hint, $comment if $comment ne '';
 
-    $string =~ s/\\"/"/g;
-    $string =~ s/\\\\/\\/g;
+    $string =~ s/(\\\d{1,3}|\\x[0-9a-f]{2}|\\u[0-9a-f]{4}|\\u\{[0-9a-f]{1,6}\}|\\[bfnrtv"'\\])/_unescape($1)/sgie;
 
     my $translated_string = &$callbackref($string, $context, join("\n", @hint), undef, $lang, $key);
 
-    $translated_string =~ s/\\/\\\\/g;
-    $translated_string =~ s/\n/\\n/sg;
-    $translated_string =~ s/"/\\"/g;
+    $translated_string =~ s/(\\[bfrv]|[\\"'\n\t])/_escape($1)/sge;
 
     return $translated_string;
 }
 
+sub _unescape {
+    my $s = shift;
+    $s eq '\\n' && return "\n";
+    $s eq '\\t' && return "\t";
+
+    if ($s =~ m/^\\(\d{1,3})$/) {
+        return chr oct $1;
+    }
+
+    if ($s =~ m/^\\x([0-9a-f]{2})$/i) {
+        return chr hex $1;
+    }
+
+    if ($s =~ m/^\\u([0-9a-f]{4})$/i) {
+        return chr hex $1;
+    }
+
+    if ($s =~ m/^\\u\{([0-9a-f]{1,6})\}$/i) {
+        return chr hex $1;
+    }
+
+    if ($s =~ m/^\\[bfrv]$/i) {
+        return $s;
+    }
+
+    $s =~ s/^\\//;
+    return $s;
+}
+
+sub _escape {
+    my $s = shift;
+    $s eq "\n" && return '\n';
+    $s eq "\t" && return '\t';
+    if ($s =~ m/^\\[bfrv]$/i) {
+        return $s;
+    }
+    return '\\'.$s;
+}
+
 1;
\ No newline at end of file
